// Led button driver software
// copyright Arthur Tollet, 2022

#include <WS2812Serial.h>
#include <Audio.h>


//init neopixel parameters
const int numled = 10;
const int pin = 10;
byte drawingMemory[numled*3];         //  3 bytes per LED
DMAMEM byte displayMemory[numled*12]; // 12 bytes per LED

WS2812Serial leds(numled, displayMemory, drawingMemory, pin, WS2812_GRB);

//define 32 bit color codes
#define RED    0xFF0000
#define PINK   0xFF1088
#define OFF    0x000000
#define BT     Serial1


//set audio parameters, from Teensy audio design utility
AudioInputI2S            i2s1;           //xy=180,111
AudioFilterStateVariable filter1;        //xy=325,101
AudioAnalyzeRMS          rms1;          //xy=327,55
AudioAmplifier           amp1;           //xy=470,93
AudioConnection          patchCord1(i2s1, 0, filter1, 0);
AudioConnection          patchCord2(filter1, 2, amp1, 0);
AudioAnalyzeFFT1024      fft1024_1;      //xy=616,102
AudioConnection          patchCord3(amp1, fft1024_1);
AudioConnection          patchCord4(amp1, rms1);

//init shared variables
volatile int selectedColor[] = {0,0,0};
volatile int state;
volatile int brt;
volatile int speed;
volatile int gate;
volatile float VUavg;
elapsedMillis fps;

void setup() {
  //general initial values
  AudioMemory(50);
  filter1.frequency(30); // filter out DC & extremely low frequencies
  amp1.gain(8.5);        // amplify sign to useful range
  leds.begin();
  Serial.begin(9600);
  BT.begin(9600);
  BT.setTimeout(20);
  state = 0;
  speed = 50;
  gate = 20;
  brt = 85;
  selectedColor[0] = 255;
  leds.setBrightness(brt);
}


// change all the LEDs in 1.5 seconds
int microsec = 1000000 / leds.numPixels();

void loop() {
  String cmd;
  int arg1,arg2,arg3 = 0;
// 3 element colors for colorWIpe
  int white[] = {255,255,255};
  int black[] = {0,0,0};
  
// Read serial. parse 1-4 arguments with spaces as delimiters.
  if (BT.available() > 0) {
    String buf = BT.readStringUntil('\n');
    //Serial.println(buf);
    int sep = buf.indexOf(" ");
    //Serial.println(sep);
    if (sep != -1){
      cmd = buf.substring(0, sep);
      int sep2 = buf.indexOf(" ", sep+1);
      //Serial.println(sep);
      arg1 = buf.substring(sep+1, sep2).toInt();
      sep = sep2;
      sep2 = buf.indexOf(" ", sep+1);
      if (sep2 != -1){
        arg2 = buf.substring(sep+1, sep2).toInt();
        arg3 = buf.substring(sep2+1).toInt();
        }
      }
    else{
      cmd = buf;
    }
  
//    Serial.print("Command: ");
//    Serial.println(cmd);
//    Serial.printf("Args: %d %d %d\n", arg1, arg2, arg3);

    char a = cmd[0];

//switch case selects program based on arg0
    switch(a){

      case('o'):
      //led off
        state = 0;
        selectedColor[0] = 0;
        selectedColor[1] = 0;
        selectedColor[2] = 0;
        BT.println("LEDs off");
        Serial.println("LEDs off");
        break;

      case('d'):
      //default values
        state = 0;
        selectedColor[0] = 255;
        selectedColor[1] = 0;
        selectedColor[2] = 0;
        brt = 85;
        speed = 50;
        gate = 20;
        leds.setBrightness(brt);
        BT.println("default values");
        Serial.println("default values");
        break;

      case('g'):
      // gold color
        state = 0;
        selectedColor[0] = 255;
        selectedColor[1] = 215;
        selectedColor[2] = 0;
        leds.setBrightness(brt);
        //print inside joke
        BT.println("16.7 Miljoner farger, och du vill ha guld...?");
        Serial.println("16.7 Miljoner farger, och du vill ha guld...?");
        break;

      case('b'):
      //set brightness
        if (arg1 < 101){
          brt = arg1*2.55;
          leds.setBrightness(brt);
          BT.printf("Setting master brightness to %d/255\n", arg1);
          Serial.printf("Setting brightness to %d/255\n", arg1);
        }
        else{
          BT.println("Please enter valid brightness value");
        }
        break;

        
      case('w'):
      //flashlight mode
        state = 0;
        selectedColor[0] = 255;
        selectedColor[1] = 255;
        selectedColor[2] = 255;
        leds.setBrightness(255);
        BT.println("Flaslight mode!");
        Serial.println("Flaslight mode!");
        break;
        

      case('c'):
      //custom color using args 1-3 as rgb values
        state = 0;
        selectedColor[0] = arg1;
        selectedColor[1] = arg2;
        selectedColor[2] = arg3;
        leds.setBrightness(brt);
        BT.printf("LEDs set to R%d G%d B%d\n", arg1,arg2,arg3);
        Serial.printf("LEDs set to R%d G%d B%d\n", arg1,arg2,arg3);
        break;

      case('q'):
      //prints instructions in terminal.
        BT.println("Welcome to BavidOS!\nSend commands through serial, in the format \n<cmd> <arg> <arg2> <arg3>\nFollowing commands are available");
        BT.println("\nStatic modes: \nd - default. Sets LEDs to red and 30% brightness \no - LEDs off \nw - flashlight, White full brightness \ng - for nar humps mouhoo\nc - custom color, give RGB values as args 1-3");
        BT.println("\nAnimations:\nr - Rainbow \nf - FFT audio visualizer\ne - Strobo\nm - VU meter\ny - Awooga :3");
        BT.println("\nModifiers (Used with Arg1):\nb - Set Brightness in % (0-100, default 34) \na - set mic gain (1-20, default 8.5) \ns - Set animation speed in % (10-100, default 50)\nx set mic gate (10-100, default 20)");
        break;

      case('r'):
      //rainbow function
        state = 1;
        leds.setBrightness(brt);
        BT.println("Rainbow mode!");
        Serial.println("Rainbow mode");
        break;

      case('f'):
      // FFT function
        leds.setBrightness(brt);
        state = 3;
        BT.println("FFT mode!");
        Serial.println("FFT mode!");
        break;

      case('y'):
      // <3
        leds.setBrightness(brt);
        state = 4;
        BT.println("Awooga :3");
        Serial.println("Awooga :3");
        break;

      case('a'):
      //set animation speed
        if (arg1 > 10 && arg1 <= 100){
          speed = arg1;
          BT.printf("Animation speed set to %d%\n", arg1);
          Serial.printf("Animation speed set to %d%\n", arg1);
        }
        else{
          BT.println("invalid speed!");
        }
        break;

      case('z'):
      //set amp gain
      if (arg1 >= 1 && arg1 <=20){
        amp1.gain(arg1);
        BT.printf("Setting amp gain to %d\n", arg1);
        Serial.printf("Setting amp gain to %d\n", arg1);
        }
      else{
          BT.println("invalid value!");
        }
        
      break;

      case('x'):
      //set gate
        if (arg1 > 5 && arg1 <= 100) {
          gate = arg1;
          BT.printf("gate set to %d%\n", arg1);
          Serial.printf("gate set to %d%\n", arg1);
          }
        else {
          BT.println("invalid value!");
          BT.println("invalid value!");
          }
        break;

      case('e'):
      // 16x strobo flash
        leds.setBrightness(255);
        for (int i=0; i<32; i++){
          if (i%2 != 0){
            colorWipe(white, 0);
          }
          else{
            colorWipe(black,0);
          }
          delay(10);
        }
        leds.setBrightness(brt);
        BT.println("Epilepsy!");
        break;

      case('m'):
      //VU meter
        state = 2;
        MicCheck();
        BT.println("VU mode!");
        Serial.println("VU mode!");
        break;

      default:
        BT.println("Invalid Command");
        Serial.println("Invalid Command");  
        break;
        
    }
  }

 switch(state){
  //looping switch case repeatedly calls function depending on state variable

  case(0):
    colorWipe(selectedColor, microsec);
    break;
  case(1):
    rainbowCycle(speed);
    break;
  case(2):
    VU();
    break;
  case(3):
    Fourier();
    break;
  case(4):
    Awooga(speed);
    break;
 }
}

void colorWipe(volatile int color[], int wait) {
  ///modified standard function. Changes color in sequence
  
    for (int i=0; i < leds.numPixels(); i++) {
      if (BT.available() == 0){
        leds.setPixel(i, color[0], color[1], color[2]);
        leds.show();
        delayMicroseconds(wait);
        }
      }
  
}

void rainbowCycle(int speed) {
  //modified standard function. cycles leds through hues
  uint16_t i, j;
  int wait = map(speed, 10, 100, 20, 1);
  
    for(j=0; j<256*2; j++) { // 1 cycles of all colors on wheel
      if (!BT.available()){
        for(i=0; i< leds.numPixels(); i++) {
          leds.setPixelColor(i, Wheel(((i * 256 / leds.numPixels()) + j) & 255));
          }
        leds.show();
        delay(wait);
        }
  }
  
}

uint32_t Wheel(byte WheelPos) {
  //stock function, that returns a 32 bit color-code based on an 8.bit number
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) {
    return leds.Color(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if(WheelPos < 170) {
    WheelPos -= 85;
    return leds.Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
  WheelPos -= 170;
  return leds.Color(WheelPos * 3, 255 - WheelPos * 3, 0);
}

void MicCheck(){
  //function reads mic values and updates average
  float vols = 0;
  int i = 0;
  while (i<60){
    if (rms1.available()){
      vols = vols + rms1.read();
      i++;
    }
    delay(1); //mystery delay, wont work without
  }
  VUavg = vols/60;
  if (VUavg < float(gate)/3000){
    VUavg = float(gate)/3000;
  }
  Serial.println(VUavg);
}

//Vu variables
volatile int count = 0;
volatile int oldcolour;
volatile float oldPeak;

// old selfwritten VU function.
//void VU() {
//  uint32_t colour;
//  //calls are spaced by 48 ms
//  if (fps%48 == 0) {
//    if (rms1.available()) {
//      if (fft1024_1.available()) {
//        // Each time new FFT data is available
//        // determines FFT bin with strongest freq.
//        float max = 0;
//        int maxval = 0;
//        for (int i = 0; i < 20; i++) {  
//          float n = fft1024_1.read(i);
//          if (n > max){
//            max = n;
//            maxval = i;
//          }
//          
//      }
//      //dampens color change to limit epilepsy
//      if (maxval-oldcolour > 2){
//              maxval = oldcolour+2;
//              }
//      else if (maxval - oldcolour < -2){
//        maxval = oldcolour-2;
//        }
//      oldcolour = maxval;
//      
//      colour = Wheel(map(maxval, 0, 20, 0, 255));
//      }
//      
//      //checks if mic gain needs recalibration
//      if (count > 8){
//        MicCheck();
//        count = 0;
//      }
//      //resets calibration variable every 640ms
//      if (fps > 64 * 10){
//        fps = 0;  
//        count = 0;
//      }
//      //Serial.println(count);
//
//      //read RMS val
//      float rawrms = rms1.read();
//      Serial.println(rawrms);
//      if (rawrms-oldPeak > 0.01){
//          rawrms = oldPeak+0.01;
//          }
//      else if (rawrms - oldPeak < 0.01){
//          rawrms = rawrms-0.01;
//          }
//      oldPeak = rawrms;
//      //dampens value change. twice the rise speed is allowed compared to fall.
//      
//      // blackouts strip if volume under gate threshold
//      if (rawrms > float(gate)/3000){
//        int monoRMS = map(rawrms, 0, VUavg*2, 0, 10);
//        //maps rms value with average in the middle of the strip
//        
//        //dampens value change. twice the rise speed is allowed compared to fall.
//        
//        //Serial.println(monoRMS);
//        for (int i = 0; i < monoRMS; i++) {
//          leds.setPixelColor(i, colour);
//          //sets n pixels based on volume to color x based on main freq.
//        }
//        for (int i = monoRMS; i < numled; i++) {
//          leds.setPixelColor(i, OFF);
//          //rest are set off
//        }
//        if (monoRMS > 8){
//            count++;
//            leds.setPixelColor(9, colour);
//            //set both top buttons to same color. increase recalibration counter as vol is too high
//          }
//        if (monoRMS > 10){
//              
//            leds.setPixelColor(8, RED);
//            leds.setPixelColor(9, RED);
//            //indicate overload by red color
//          }
//        if (monoRMS < 2){
//          count++;
//          //increase recalibration counter as vol is too low 
//        }
//        
//        leds.show();
//        }
//        else{
//          for (int i = 0; i < numled; i++) {
//          leds.setPixelColor(i, OFF);
//          oldPeak = 0;
//          oldcolour = 0;
//          //reset vol and color on off state
//        }
//        leds.show();
//        }
//      }
//      
//  }
//}

//New VU function

#define TOP       (numled + 2) // Allow dot to go slightly off scale
#define PEAK_FALL 60  // Rate of peak falling dot
#define SAMPLES   80  // Length of buffer for dynamic level adjustment

byte
  peak      = 0,      // Used for falling dot
  dotCount  = 0,      // Frame counter for delaying dot-falling speed
  volCount  = 0;      // Frame counter for storing past volume data
int
  vol[SAMPLES],       // Collection of prior volume samples
  lvl       = 20,      // Current "dampened" audio level
  minLvlAvg = 0,      // For dynamic adjustment of graph low & high
  maxLvlAvg = 512;

void VU() {
    uint8_t  i;
    uint16_t minLvl, maxLvl;
    int      n, height;
  
  
  if(rms1.available()){
    
    n   = rms1.read()*1000;                  // Raw reading from mic
    lvl = ((lvl * 7) + n) >> 3;    // "Dampened" reading (else looks twitchy)
  
    // Calculate bar height based on dynamic min/max levels (fixed point):
    height = TOP * (lvl - minLvlAvg) / (long)(maxLvlAvg - minLvlAvg);
  
    if(height < 0L)       height = 0;      // Clip output
    else if(height > TOP) height = TOP;
    if(height > peak)     peak   = height; // Keep 'peak' dot at top
  
  
    // Color pixels based on rainbow gradient
    for(i=0; i<numled; i++) {
      if(i >= height)               leds.setPixelColor(i,   0,   0, 0);
      else leds.setPixelColor(i,Wheel(map(i,0,leds.numPixels()-1,30,150)));
  
    }
  
  
  
    // Draw peak dot 
    if(peak > 0 && peak <= numled-1) leds.setPixelColor(peak,Wheel(map(peak,0,leds.numPixels()-1,30,150)));
  
     leds.show(); // Update strip
  
  // Every few frames, make the peak pixel drop by 1:
  
      if(++dotCount >= PEAK_FALL) { //fall rate
  
        if(peak > 0) peak--;
        dotCount = 0;
      }
  
  
  
    vol[volCount] = n;                      // Save sample for dynamic leveling
    if(++volCount >= SAMPLES) volCount = 0; // Advance/rollover sample counter
  
    // Get volume range of prior frames
    minLvl = maxLvl = vol[0];
    for(i=1; i<SAMPLES; i++) {
      if(vol[i] < minLvl)      minLvl = vol[i];
      else if(vol[i] > maxLvl) maxLvl = vol[i];
    }
    // minLvl and maxLvl indicate the volume range over prior frames, used
    // for vertically scaling the output graph (so it looks interesting
    // regardless of volume level).  If they're too close together though
    // (e.g. at very low volume levels) the graph becomes super coarse
    // and 'jumpy'...so keep some minimum distance between them (this
    // also lets the graph go to zero when no sound is playing):
    if((maxLvl - minLvl) < TOP) maxLvl = minLvl + TOP;
    minLvlAvg = (minLvlAvg * 63 + minLvl) >> 6; // Dampen min/max levels
    maxLvlAvg = (maxLvlAvg * 63 + maxLvl) >> 6; // (fake rolling average)
  
  }
}

// FFT visualizer. Pretty chaotic atm
void Fourier() {
  if (fft1024_1.available()) {
    // Each time new FFT data is available
    // loop through FFT bins, read both minimum and maximum frequencies.
    float min = 0;
    float max = 0;
    for (int i = 0; i < 512; i++) {
      
      float n = fft1024_1.read(i);
      if (n > max){
        max = n;
      }
      else if (n <= min){
        min = n;
      }
    }
    for (int i = 0; i < 10; i++) {
      // assign first 20 bins to buttons, in summed pairs.
      float n = fft1024_1.read(i*2, i*2+1);
      float g = float(gate)/10000;
      if (n >= g) {
        int colour = Wheel(map(n, min, max, 0, 255));
        //determine colour on range from weakest to strongest freq
        leds.setPixelColor(i, colour);
        Serial.print(n);
        Serial.print(",");
      } else {
        leds.setPixelColor(i, OFF);
        Serial.print("-");
      }
    }
    leds.show();
    Serial.println();
  }
}


//mood light :3
void Awooga(int speed) {
  int wait = map(speed, 10, 100, 40, 10);
  int i;
  //set pink color
  for (i = 0; i < numled; i++) {
    leds.setPixelColor(i, PINK);
    }

  //breathing animation
  for (i = brt; i > 10 && !BT.available(); i--) {
    leds.setBrightness(i);
    leds.show();
    delay(wait);
    }
  for (i = 10; i < brt && !BT.available(); i++) {
    leds.setBrightness(i);
    leds.show();
    delay(wait);
    }
}
